@@ -5,7 +5,7 @@ use warnings;
 
 use parent qw(IO::Async::Notifier);
 
-our $VERSION = '0.031';
+our $VERSION = '1.000';
 
 =head1 NAME
 
@@ -13,7 +13,7 @@ Net::Async::AMQP - provides client interface to AMQP using L<IO::Async>
 
 =head1 VERSION
 
-version 0.031
+version 1.000
 
 =head1 SYNOPSIS
 
@@ -30,12 +30,14 @@ version 0.031
 =head1 DESCRIPTION
 
 Does AMQP things. Note that the API may change before the stable 1.000
-release - L</SEE ALSO> has some alternative modules if you're looking for
-something that has been around for longer.
+release - L</ALTERNATIVE AMQP IMPLEMENTATIONS> are listed below if you want
+to evaluate other options.
 
 If you want a higher-level API which manages channels and connections, try
 L<Net::Async::AMQP::ConnectionManager>.
 
+Examples are in the C<examples/> directory.
+
 =head2 AMQP support
 
 The following AMQP features are supported:
@@ -87,6 +89,9 @@ if you want any of these:
 
 =back
 
+This implementation is designed to handle many simultaneous channels and connections. If you just want a
+single consumer/publisher, one of the librabbitmq-c implementations may be sufficient.
+
 =cut
 
 use Net::AMQP;
@@ -378,22 +383,6 @@ Called whenever there's data available to be read.
 
 sub on_read {
 	my ($self, $stream, $buffref, $eof) = @_;
-	# Frame dumping support - not that useful yet, so it's disabled
-	if(0) {
-		my $mem = $$buffref;
-		$self->dump_frame($mem);
-		my $idx = 0;
-		while(length $mem) {
-			my $hex = join ' ', unpack 'H2'x16, my $bytes = substr $mem, 0, 16, '';
-			substr $hex, 8 * 3, 0, '  ';
-			my $asc = join '', map /([[:print:]])/ ? $1 : '.', split //, $bytes;
-			substr $asc, 8, 0, ' ';
-			printf "%8d:  %-52.52s %s\n", $idx, $hex, $asc;
-			$idx += length($asc);
-		}
-		print "\n";
-		$self->debug_printf("At EOF") if $eof;
-	}
 
 	$self->last_frame_time(Time::HiRes::time);
 
@@ -414,11 +403,11 @@ sub on_closed {
 	my $reason = shift // 'unknown';
 	$self->debug_printf("Connection closed [%s]", $reason);
 
-	for my $ch (values %{$self->{channel_by_id}}) {
+	for my $ch (grep $_, values %{$self->{channel_by_id}}) {
 		$ch->bus->invoke_event(
 			'close',
-			# code    => 999,
-			message => 'Connection closed: ' . $reason,
+			code    => undef,
+			reason  => 'Connection closed: ' . $reason,
 		);
 		$self->channel_closed($ch->id);
 	}
@@ -429,7 +418,8 @@ sub on_closed {
 
 	$self->stream->close if $self->stream;
 	for (qw(stream heartbeat_send_timer heartbeat_receive_timer)) {
-		$self->remove_child(delete $self->{$_}) if $self->{$_};
+		$self->debug_printf("Remove child %s", $_);
+		(delete $self->{$_})->remove_from_parent if $self->{$_};
 	}
 	$self->bus->invoke_event(close => $reason)
 }
@@ -691,7 +681,7 @@ sub close {
 	my $frame = Net::AMQP::Frame::Method->new(
 		method_frame => Net::AMQP::Protocol::Connection::Close->new(
 			reply_code => $args{code} // 320,
-			reply_text => $args{text} // 'Request connection close',
+			reply_text => $args{reason} // 'Request connection close',
 		),
 	);
 	$self->push_pending(
@@ -749,6 +739,11 @@ sub next_pending {
 	my ($self, $type, $frame) = @_;
 	$self->debug_printf("Check next pending for %s", $type);
 
+	if($type eq 'Connection::Close') {
+		$self->on_closed($frame->method_frame->reply_text);
+		return $self;
+	}
+
 	if(my $next = shift @{$self->{pending}{$type} || []}) {
 		# We have a registered handler for this frame type. This usually
 		# means that we've sent a frame and are awaiting a response.
@@ -1098,8 +1093,8 @@ sub process_frame {
 		# A peer that receives an invalid heartbeat frame MUST raise a connection
 		# exception with reply code 501 (frame error)
 		$self->close(
-			reply_code => 501,
-			reply_text => 'Frame error - heartbeat should have channel 0'
+			code   => 501,
+			reason => 'Frame error - heartbeat should have channel 0'
 		) if $frame->channel;
 
 		return $self;
@@ -1323,6 +1318,41 @@ If we receive an unsolicited frame from the server this event will be raised:
   }
  )
 
+=head1 ALTERNATIVE AMQP IMPLEMENTATIONS
+
+As usual there's a few other options:
+
+=over 4
+
+=item * L<Net::RabbitMQ> - basic bindings for librabbitmq
+
+=item * L<Net::AMQP::RabbitMQ> - a fork of Net::RabbitMQ ("uses a newer version of librabbitmq and fixes some bugs")
+
+=item * L<Net::RabbitMQ::Client> - yet another set of bindings for librabbitmq, includes a "Simple" wrapper implementation as well
+
+=item * L<Crixa> - wrapper over L<Net::AMQP::RabbitMQ>
+
+=item * L<Net::RabbitMQ::Channel> - another wrapper around L<Net::RabbitMQ>
+
+=item * L<Net::RabbitMQ::Java> - uses the official Java client via L<Inline::Java>
+
+=item * L<Net::Thumper> - sync client based on L<Net::AMQP>, no channel support
+
+=item * L<POE::Component::Client::AMQP> - provides a L<POE> component based on L<Net::AMQP>.
+
+=item * L<AnyEvent::RabbitMQ> - uses L<Net::AMQP> to provide an L<AnyEvent> implementation.
+
+=item * L<Net::RabbitFoot> - wrapper around L<AnyEvent::RabbitMQ>
+
+=item * L<AnyMQ::AMQP> - provides AMQP support for the L<AnyMQ> abstraction, via L<AnyEvent::RabbitMQ>
+
+=back
+
+Modules based on librabbitmq-c are probably fine for simple sync tasks, but I wouldn't recommend them for
+any async work. In contrast, the L<Net::AMQP> protocol module generates all the classes and methods directly
+from the AMQP spec, so it's an excellent base on which to develop the transport module (as in the case of
+L<Net::Async::AMQP>).
+
 =head1 SEE ALSO
 
 =over 4
@@ -1330,13 +1360,9 @@ If we receive an unsolicited frame from the server this event will be raised:
 =item * L<Net::AMQP> - this does all the hard work of converting the XML protocol
 specification into appropriate Perl methods and classes.
 
-=item * L<Net::RabbitMQ> - probably bindings for librabbitmq
+=item * L<Net::RabbitMQ::Management::API> - doesn't do AMQP, but provides sync (LWP-based) access to RabbitMQ's HTTP API
 
-=item * L<Net::AMQP::RabbitMQ> - librabbitmq support
-
-=item * L<POE::Component::Client::AMQP> - POE equivalent of this module
-
-=item * L<AnyEvent::RabbitMQ>
+=item * L<Test::Net::RabbitMQ> - provides a basic server implementation for testing
 
 =back
 
