@@ -1,5 +1,5 @@
 package Net::Async::AMQP::Channel;
-$Net::Async::AMQP::Channel::VERSION = '0.031';
+$Net::Async::AMQP::Channel::VERSION = '1.000';
 use strict;
 use warnings;
 
@@ -11,7 +11,7 @@ Net::Async::AMQP::Channel - represents a single channel in an MQ connection
 
 =head1 VERSION
 
-version 0.031
+version 1.000
 
 =head1 SYNOPSIS
 
@@ -98,19 +98,40 @@ sub confirm_mode {
 	my $self = shift;
 	my %args = @_;
 	$self->debug_printf("Enabling confirm mode");
+	die "already requested confirm_mode for this channel" if $self->{confirm_mode};
 
-	my $f = $self->loop->new_future;
+	my $f = $self->loop->new_future->set_label("set confirm_mode on channel " . $self->id);
+	$self->{delivery_tag} = 0;
+	$self->{confirm_mode} = $f;
+	my $nowait = $self->nowait_from_args(%args);
 	my $frame = Net::AMQP::Frame::Method->new(
 		method_frame => Net::AMQP::Protocol::Confirm::Select->new(
-			nowait      => 0,
+			nowait => $nowait,
 		)
 	);
-	$self->push_pending(
-		'Confirm::SelectOk' => [ $f, $self ]
-	);
-	$self->closure_protection($f);
+
+	# No-wait mode means we don't expect the SelectOk frame back
+	if($nowait) {
+		$f->done
+	} else {
+		$self->closure_protection($f);
+	}
 	$self->send_frame($frame);
-	return $f;
+	return $f->transform(done => sub { $self });
+}
+
+=head2 nowait_from_args
+
+If we have a C<wait> argument, then return the inverse of that.
+
+Otherwise, return zero.
+
+=cut
+
+sub nowait_from_args {
+	my ($self, %args) = @_;
+	return 0 unless exists $args{wait};
+	return $args{wait} ? 0 : 1;
 }
 
 =head2 exchange_declare
@@ -134,9 +155,10 @@ sub exchange_declare {
 	die "No exchange specified" unless exists $args{exchange};
 	die "No exchange type specified" unless exists $args{type};
 
+	$args{exchange} //= '';
 	$self->debug_printf("Declaring exchange [%s]", $args{exchange});
 
-	my $f = $self->loop->new_future;
+	my $f = $self->loop->new_future->set_label("declare exchange [" . $args{exchange} . "]");
 	my $frame = Net::AMQP::Frame::Method->new(
 		method_frame => Net::AMQP::Protocol::Exchange::Declare->new(
 			exchange    => Net::AMQP::Value::String->new($args{exchange}),
@@ -171,7 +193,7 @@ sub exchange_bind {
 
 	$self->debug_printf("Binding exchange [%s] to [%s] with rkey [%s]", $args{source}, $args{destination}, $args{routing_key});
 
-	my $f = $self->loop->new_future;
+	my $f = $self->loop->new_future->set_label("bind exchange [" . $args{source} . "] to [" . $args{destination} . "]" . (exists $args{routing_key} ? (" rkey [" . $args{routing_key} . "]") : ""));
 	my $frame = Net::AMQP::Frame::Method->new(
 		method_frame => Net::AMQP::Protocol::Exchange::Bind->new(
 			source      => Net::AMQP::Value::String->new($args{source}),
@@ -190,12 +212,12 @@ sub exchange_bind {
 
 =head2 queue_declare
 
-Returns a L<Future> which will resolve with the
-new L<Net::Async::AMQP::Queue> instance once complete.
+Returns a L<Future> which will resolve with the new L<Net::Async::AMQP::Queue> instance,
+the number of messages in the queue, and the number of consumers.
 
  $ch->queue_declare(
   queue      => 'some_queue',
- ) ==> $q
+ ) ==> ($q, $message_count, $consumer_count)
 
 =cut
 
@@ -205,8 +227,8 @@ sub queue_declare {
 	die "No queue specified" unless defined $args{queue};
 
 	$self->future->then(sub {
-		my $f = $self->loop->new_future;
-		my $ready = $self->loop->new_future;
+		my $f = $self->loop->new_future->set_label("declare queue [" . $args{queue} . "]");
+		my $ready = $self->loop->new_future->set_label("queue readiness for [" . $args{queue} . "]");
 		my $q = Net::Async::AMQP::Queue->new(
 			amqp    => $self->amqp,
 			future  => $ready,
@@ -233,8 +255,10 @@ sub queue_declare {
 				my ($amqp, $frame) = @_;
 				my $method_frame = $frame->method_frame;
 				$q->queue_name($method_frame->queue);
+				my $messages = $method_frame->message_count;
+				my $consumer_count = $method_frame->consumer_count;
 				$ready->done() unless $ready->is_ready;
-				$f->done($q) unless $f->is_ready;
+				$f->done($q, $messages, $consumer_count) unless $f->is_ready;
 			}
 		);
 		$self->closure_protection($f);
@@ -243,6 +267,8 @@ sub queue_declare {
 	})
 }
 
+sub next_dtag { ++shift->{delivery_tag} }
+
 =head2 publish
 
 Publishes a message on this channel.
@@ -290,38 +316,12 @@ sub publish {
 	die "no exchange" unless exists $args{exchange};
 
 	$self->future->then(sub {
-		my $f = $self->loop->new_future;
-		{ # When publishing a message, we should expect either an ACK, or a return.
-		  # Since these are mutually exclusive, we also need to remove the pending
-		  # handler for the opposing event once one event has been received. Note
-		  # that this crosslinking gives us an unfortunate cycle which we resolve
-		  # by weakening the opposite handler once we've removed it.
-			my $return;
-			my $ack = sub {
-				my ($amqp, $frame) = @_;
-				my $method_frame = $frame->method_frame;
-				$self->remove_pending('Basic::Return' => $return);
-				$f->done unless $f->is_ready;
-				weaken $return;
-			};
-			$return = sub {
-				my ($amqp, $frame) = @_;
-				my $method_frame = $frame->method_frame;
-				# $self->remove_pending('Basic::Ack' => $ack);
-				$f->fail(
-					$method_frame->reply_text,
-					code     => $method_frame->reply_code,
-					exchange => $method_frame->exchange,
-					rkey     => $method_frame->routing_key
-				) unless $f->is_ready;
-				weaken $ack;
-			};
-			$self->push_pending(
-				'Basic::Return' => $return,
-			);
-			$self->push_pending(
-				'Basic::Ack' => $ack,
-			);
+		my $f = $self->loop->new_future->set_label("publish on [" . $args{exchange} . "]");
+		my $dtag = $self->next_dtag;
+		if($self->{confirm_mode}) {
+			push @{$self->{published}}, [ $dtag => $f ];
+		} else {
+			$f->done;
 		}
 
 		my @frames = $self->amqp->split_payload(
@@ -334,18 +334,8 @@ sub publish {
 			content_type     => $args{content_type} // 'application/binary',
 			content_encoding => $args{content_encoding},
 			timestamp        => $args{timestamp} // time,
-			type             => Net::AMQP::Value::String->new($args{type}),
+			type             => Net::AMQP::Value::String->new($args{type} // ''),
 			user_id          => $self->amqp->user,
-			no_ack           => (
-				# inverse flags are a pain... we default to ACK mode.
-				# If the ack parameter is given, this overrides the
-				# setting - so ack => 1 means no_ack => 0, ack => 0 means no_ack => 1,
-				# and lack of ack means no_ack => 0.
-				exists $args{ack}
-				? (
-					$args{ack} ? 0 : 1
-				) : 0
-			),
 #            headers          => {
 #                type => $args{type},
 #            },
@@ -394,7 +384,7 @@ sub qos {
 	my %args = @_;
 
 	$self->future->then(sub {
-		my $f = $self->loop->new_future;
+		my $f = $self->loop->new_future->set_label("set qos count " . ($args{prefetch_count} // 0) . " size " . ($args{prefetch_size} // 0));
 		my $channel = $self->id;
 		$self->push_pending(
 			'Basic::QosOk' => [ $f, $self ],
@@ -465,6 +455,7 @@ sub on_close {
 	my ($self, $frame) = @_;
 
 	$self->{is_closed} = 1;
+	$self->{future} = Future->fail('closed');
 
 	# ACK the close first - we have to send a close-ok
 	# before it's legal to reopen this channel ID
@@ -479,17 +470,28 @@ sub on_close {
 					)
 				)
 			)
-		)->then(sub {
+		)->on_ready(sub {
+			# Any remaining consumers need to be cancelled at this point
+			$self->bus->invoke_event(
+				'cancel',
+				ctag => $_,
+			) for keys %{$self->{consumer_tags}};
+			$self->{consumer_tags} = {};
+
+			$_->fail('channel closed') for grep !$_->is_ready, map $_->[1], @{$self->{published}};
+			$self->{published} = [];
+
 			# It's important that the MQ instance knows
 			# about the channel closure first before we
 			# go ahead and dispatch events, since any
 			# subscribed handlers might go ahead and
 			# attempt to open the channel again immediately.
+
 			$self->amqp->channel_closed($self->id);
 			$self->bus->invoke_event(
 				'close',
 				code => $frame->reply_code,
-				message => $frame->reply_text,
+				reason => $frame->reply_text,
 			);
 			Future->done
 		})
@@ -536,8 +538,9 @@ sub close {
 	return Future->done if $self->{closing} or !$self->loop;
 
 	$self->{closing} = 1;
+	$self->{future} = Future->fail('closing');
 
-	my $f = $self->loop->new_future;
+	my $f = $self->loop->new_future->set_label("Close channel " . $self->id);
 	my $frame = Net::AMQP::Frame::Method->new(
 		method_frame => Net::AMQP::Protocol::Channel::Close->new(
 			reply_code  => $args{code} // 404,
@@ -662,7 +665,10 @@ sub next_pending {
 
 	return $self unless $frame->can('method_frame') && (my $method_frame = $frame->method_frame);
 	my $type = amqp_frame_type($frame);
-	if($type eq 'Basic::CancelOk') {
+	if($type eq 'Basic::ConsumeOk') {
+		my $ctag = $method_frame->consumer_tag;
+		$self->{consumer_tags}{$ctag} = 1;
+	} elsif($type eq 'Basic::Cancel') {
 		my ($ctag) = ($method_frame->consumer_tag);
 		$self->debug_printf("Cancel $ctag");
 		$self->bus->invoke_event(
@@ -674,6 +680,7 @@ sub next_pending {
 		$self->bus->invoke_event(
 			listener_stop => $ctag
 		);
+		delete $self->{consumer_tags}{$ctag};
 	}
 
 	# Message delivery, part 3: The "Deliver" message.
@@ -691,13 +698,82 @@ sub next_pending {
 	}
 
 	if($type eq 'Channel::Close') {
-		$self->debug_printf("Channel was %d, calling close", $frame->channel);
+		$self->debug_printf(
+			"Channel was %d, calling close - code %d, text '%s', class:method %d:%d",
+			$frame->channel,
+			$method_frame->reply_code,
+			$method_frame->reply_text,
+			$method_frame->class_id,
+			$method_frame->method_id,
+		);
 		$self->on_close(
 			$method_frame
 		);
 		return $self;
 	}
 
+	# Confirm mode => mark pending task as done
+	if($type eq 'Confirm::SelectOk') {
+		$self->debug_printf("Confirm mode enabled");
+		$self->{confirm_mode}->done;
+		return $self;
+	} elsif($type eq 'Basic::Ack') {
+		shift @{$self->{pending}{'Basic::Return'} || []};
+		eval {
+			my @msg = $self->extract_published($method_frame->delivery_tag, $method_frame->multiple);
+			$self->debug_printf("received ack for %d messages", 0 + @msg);
+			$_->done for grep !$_->is_ready, map $_->[1], @msg;
+			1
+		} or do {
+			my $err = $@;
+			$self->debug_printf("error retrieving messages for ack - %s", $err);
+			$self->close(
+				code => 406,
+				text => $err
+			);
+		};
+		return $self;
+	} elsif($type eq 'Basic::Nack') {
+		shift @{$self->{pending}{'Basic::Return'} || []};
+		eval {
+			my @msg = $self->extract_published($method_frame->delivery_tag, $method_frame->multiple);
+			$self->debug_printf("received nack for %d messages", 0 + @msg);
+			$_->fail('nack') for grep !$_->is_ready, map $_->[1], @msg;
+			1
+		} or do {
+			my $err = $@;
+			$self->debug_printf("error retrieving messages for nack - %s", $err);
+			$self->close(
+				code => 406,
+				text => $err
+			);
+		};
+		return $self;
+	} elsif($type eq 'Basic::Return') {
+		# Basic::Return would always be for the first unacked message in our publish
+		# queue... except when we don't have publisher confirms, in which case... uh...
+		# okay in that case we'd just raise an event maybe
+		if($self->{confirm_mode}) {
+			$self->debug_printf("basic::return in confirm mode");
+			my $f = $self->{published}[0][1];
+			$f->fail(
+				$method_frame->reply_text,
+				code     => $method_frame->reply_code,
+				exchange => $method_frame->exchange,
+				rkey     => $method_frame->routing_key
+			) if $f && !$f->is_ready;
+		} else {
+			$self->debug_printf("basic::return in normal mode");
+			$self->bus->invoke_event(
+				return => $method_frame->reply_text,
+				code     => $method_frame->reply_code,
+				exchange => $method_frame->exchange,
+				rkey     => $method_frame->routing_key
+			)
+		}
+		return $self;
+	}
+
 	if(my $next = shift @{$self->{pending}{$type} || []}) {
 		# We have a registered handler for this frame type. This usually
 		# means that we've sent a message and are awaiting a response.
@@ -716,6 +792,31 @@ sub next_pending {
 	return $self;
 }
 
+sub extract_published {
+	my ($self, $dtag, $multiple) = @_;
+	if($multiple) {
+		my @msg;
+		while(@{$self->{published}}) {
+			my $msg = shift @{$self->{published}};
+			# Nonzero dtag means "up to and including this message"
+			die 'ack for dtag ' . $dtag . ' but our earliest message is ' . $msg->[0] if $dtag && $dtag < $msg->[0];
+
+			# with dtag=0 that's "everything you've got". this is fundamentally
+			# flawed since the server may not have even received the most recent
+			# items... might work if the server *always* uses dtag=0, I guess
+			push @msg, $msg;
+			last if $dtag == $msg->[0];
+		}
+		return @msg;
+	}
+	# Single-ack *probably* handles things in order, but the spec does not seem
+	# to mandate this - better to be safe
+	for my $idx (0..$#{$self->{published}}) {
+		return splice @{$self->{published}}, $idx, 1 if $self->{published}[$idx][0] == $dtag;
+	}
+	die 'ack for dtag ' . $dtag . ' but not found in our pending list (we had ' . @{$self->{published}} . ' pending messages)';
+}
+
 =head1 METHODS - Accessors
 
 =cut
@@ -784,35 +885,39 @@ Takes a L<Future>, returns a L<Future> (probably the same one).
 sub closure_protection {
 	my ($self, $f) = @_;
 	unless($f) {
-		$self->debug_printf("Closure protection requested for future which has already disappeared");
+		$self->debug_printf("Closure protection requested on channel %d for future which has already disappeared", $self->id);
 		return Future->fail(closed => 'future has already been released');
 	}
 
 	# No sense in proceeding if the Future has already completed
-	return $f if $f->is_ready;
+	if($f->is_ready) {
+		$self->debug_printf("Closure protection requested for future %s on channel %d which has already compelted", $f->label, $self->id);
+		return $f;
+	}
 
+	my $id = $self->id;
 	my @ev;
 	my $bus = $self->bus;
 	$bus->subscribe_to_event(
 		@ev = (close => sub {
 			my ($ev, %args) = @_;
-			$self->debug_printf("Closed channel, code %d, reason: %s", $args{code}, $args{reason});
-			unless($f) {
+			$self->debug_printf("Closure protection engaging for %s on channel %d, code %s, reason: %s", ($f ? $f->label : "(future which no longer exists)"), $id, $args{code} // '(none)', $args{reason});
+			if($f) {
+				$f->fail($args{reason}, 'amqp', $args{code}) unless $f->is_ready;
+			} else {
 				$self->debug_printf("Future has disappeared already, not marking as failed");
-				# We should have unsubscribed already, but do this just in case.
-				splice @ev;
-				eval { $ev->unsubscribe; };
-				return;
 			}
-
-			$f->fail(closed => %args) unless $f->is_ready;
+			# We should have unsubscribed already, but do this just in case.
+			splice @ev;
+			eval { $ev->unsubscribe; };
 		})
 	);
 
 	# Use return value from ->on_ready, since we may clear $f immediately if the future is already
 	# marked as ready.
 	$f->on_ready(sub {
-		$bus->unsubscribe_from_event(splice @ev);
+		$self->debug_printf("Future %s on channel %d is ready, disengaging closure protection", ($f ? $f->label : "(future which no longer exists)"), $id);
+		eval { $bus->unsubscribe_from_event(splice @ev); };
 		undef $f;
 	});
 }
