@@ -1,5 +1,5 @@
 package Net::Async::AMQP::ConnectionManager::Channel;
-$Net::Async::AMQP::ConnectionManager::Channel::VERSION = '0.031';
+$Net::Async::AMQP::ConnectionManager::Channel::VERSION = '1.000';
 use strict;
 use warnings;
 
@@ -9,7 +9,7 @@ Net::Async::AMQP::ConnectionManager::Channel - channel proxy object
 
 =head1 VERSION
 
-version 0.031
+version 1.000
 
 =cut
 
@@ -45,8 +45,9 @@ sub new {
 	);
 
 	$self->{cleanup}{events} = sub {
-		shift->bus->unsubscribe_from_event(splice @ev);
-		Future->wrap;
+		my $ch = shift;
+		$ch->bus->unsubscribe_from_event(splice @ev);
+		Future->done;
 	};
 	$self
 }
@@ -102,14 +103,14 @@ sub queue_declare {
 	my ($self, %args) = @_;
 	$self->channel->queue_declare(%args)->transform(
 		done => sub {
-			my ($q) = @_;
+			my ($q, @extra) = @_;
 			# Ensure that this wrapped channel is used
 			# as the stored channel value. This means
 			# the channel holds the queue, we hold a weakref
 			# to the channel, and the queue holds a strong
 			# ref to our channel wrapper.
 			$q->configure(channel => $self);
-			$q
+			return ($q, @extra);
 		}
 	)
 }
@@ -195,13 +196,15 @@ sub DESTROY {
 		return $conman->release_channel($ch);
 	}
 
+	return unless $self->{channel};
+
 	my $f;
 	my $cleanup = delete $self->{cleanup};
 	$f = (
 		fmap_void {
 			my $k = shift;
 			my $task = delete $cleanup->{$k};
-			$task ? $task->($self) : Future->wrap
+			$task ? $task->($self) : Future->done
 		} foreach => [
 			sort keys %$cleanup
 		]
