@@ -1,5 +1,5 @@
 package Net::Async::AMQP::Queue;
-$Net::Async::AMQP::Queue::VERSION = '0.031';
+$Net::Async::AMQP::Queue::VERSION = '1.000';
 use strict;
 use warnings;
 
@@ -9,7 +9,7 @@ Net::Async::AMQP::Queue - deal with queue-specific functionality
 
 =head1 VERSION
 
-version 0.031
+version 1.000
 
 =cut
 
@@ -65,7 +65,7 @@ sub listen {
 
     # Attempt to bind after we've successfully declared the exchange.
     retain_future($self->future->then(sub {
-        my $f = $ch->loop->new_future;
+        my $f = $ch->loop->new_future->set_label("listen on " . $self->queue_name);
         $ch->debug_printf("Attempting to listen for events on queue [%s]", $self->queue_name);
 
         my $frame = Net::AMQP::Protocol::Basic::Consume->new(
@@ -142,7 +142,7 @@ sub cancel {
 
     # Attempt to bind after we've successfully declared the exchange.
 	retain_future($self->future->then(sub {
-		my $f = $ch->loop->new_future;
+		my $f = $ch->loop->new_future->set_label("cancel ctag " . $ctag);
 		$ch->debug_printf("Attempting to cancel consumer [%s]", $ctag);
 
 		my $frame = Net::AMQP::Protocol::Basic::Cancel->new(
@@ -165,6 +165,104 @@ sub cancel {
 	}));
 }
 
+=head2 consumer
+
+Similar to L</listen>, but applies the event handlers so you can just provide an C<on_message> callback.
+
+Takes the following extra named parameters:
+
+=over 4
+
+=item * on_message - callback for message handling
+
+=item * on_cancel - will be called if the consumer is cancelled (either by the server or client)
+
+=back
+
+For server consumer cancellation notification, you'll need consumer_cancel_notifications:
+
+ $mq->connect(
+  ...
+  client_properties => {
+   capabilities => {
+    'consumer_cancel_notify' => Net::AMQP::Value->true
+   },
+  },
+ )
+
+The on_message callback receives the following named parameters:
+
+=over 4
+
+=item * type
+
+=item * payload
+
+=item * consumer_tag
+
+=item * delivery_tag
+
+=item * routing_key 
+
+=back
+
+See C<examples/alternative-consumer.pl> for a usage example.
+
+=cut
+
+sub consumer {
+    my ($self, %args) = @_;
+
+	my $ch = (delete $args{channel}) // die "No channel";
+	my $ctag = (delete $args{consumer_tag}) // '';
+	my $on_message = delete $args{on_message} || sub { };
+	my $on_cancel = delete $args{on_cancel} || sub { };
+	my @ev;
+	$ch->bus->subscribe_to_event(
+		@ev = (
+			# Deliver any matching messages to our callback
+			message => sub {
+				my ($ev, $type, $payload, $incoming_ctag, $dtag, $rkey) = @_;
+				return unless $incoming_ctag eq $ctag;
+				$on_message->(
+					type => $type,
+					payload => $payload,
+					consumer_tag => $ctag,
+					delivery_tag => $dtag,
+					routing_key => $rkey
+				);
+			},
+			# Drop event handlers and call cancellation callback on cancel
+			cancel => sub {
+				my ($ev, %args) = @_;
+				return unless $args{ctag} eq $ctag;
+				# Avoid potential race between ->cancel and the consumer future being cancelled
+				eval {
+					$ch->bus->unsubscribe_from_events(@ev);
+				};
+				$on_cancel->(
+					consumer_tag => $ctag
+				);
+			}
+		)
+	);
+	$self->listen(
+		%args,
+		channel => $ch,
+		consumer_tag => $ctag,
+	)->on_done(sub {
+		$ctag = $_[1];
+	})->on_ready(sub {
+		return if shift->is_done;
+		eval {
+			$ch->bus->unsubscribe_from_events(@ev);
+		};
+		$on_cancel->(
+			consumer_tag => $ctag
+		)
+	});
+}
+
 =head2 bind_exchange
 
 Binds this queue to an exchange.
@@ -203,7 +301,7 @@ sub bind_exchange {
 
     # Attempt to bind after we've successfully declared the exchange.
 	retain_future($self->future->then(sub {
-		my $f = $ch->loop->new_future;
+		my $f = $ch->loop->new_future->set_label(sprintf "bind exchange [%s] to exchange [%s] with rkey [%s]", $self->queue_name, $args{exchange}, $args{routing_key} // '(none)');
 		$ch->debug_printf("Binding queue [%s] to exchange [%s] with rkey [%s]", $self->queue_name, $args{exchange}, $args{routing_key} // '(none)');
 
 		my $frame = Net::AMQP::Frame::Method->new(
@@ -262,7 +360,7 @@ sub unbind_exchange {
 
     # Attempt to unbind after we've successfully declared the exchange.
 	retain_future($self->future->then(sub {
-		my $f = $ch->loop->new_future;
+		my $f = $ch->loop->new_future->set_label("unbind exchange [%s] from exchange [%s] with rkey [%s]", $self->queue_name, $args{exchange}, $args{routing_key} // '(none)');
 		$ch->debug_printf("Unbinding queue [%s] from exchange [%s] with rkey [%s]", $self->queue_name, $args{exchange}, $args{routing_key} // '(none)');
 
 		my $frame = Net::AMQP::Frame::Method->new(
@@ -314,7 +412,7 @@ sub delete : method {
 	my $ch = delete $args{channel} or die "No channel provided";
 
 	retain_future($self->future->then(sub {
-		my $f = $ch->loop->new_future;
+		my $f = $ch->loop->new_future->set_label("delete " . $self->queue_name);
 		$ch->debug_printf("Deleting queue [%s]", $self->queue_name);
 
 		my $frame = Net::AMQP::Frame::Method->new(
